;***************************************************************************
;
; iButton controlled lock
;
; Created by Nicholas Redgrave 15th December 2005
;
; Chunks of code (c) M.Prinke (Modified by Nicholas Redgrave)
;
;
;***************************************************************************

;***** DIRECTIVES *****
  EXPAND
  LIST b=2
  RADIX dec

;***** PROCESSOR DECLARATION & CONFIGURATION *****

  PROCESSOR p16f84

  ; embed Configuration Data within .asm File.

  __CONFIG  _CP_OFF & _WDT_OFF & _PWRTE_OFF & _HS_OSC


  org 0


;***** INCLUDE FILES *****

  include /usr/share/gputils/header/p16f84.inc



;***** CONSTANT DECLARATION *****

SID     equ 3     ;number of stored iButton IDs

BLACK   equ 0
YELLOW  equ 3
RED     equ 0
GREEN   equ 1
LOCK    equ 2
IB      equ 1
RROM    equ 0x33  ;iButton Read ROM cmd
TAB_PG  equ 1     ;page of lut begin


;***** VARIABLE DECLARATION *****

  ORG 0x0C    ;Base address of user file registers

bdata   RES 1 ;iButton databyte
crc     RES 1 ;crc
crcb    RES 1 ;byte to put into crc
bcount  RES 1 ;bitcounter
tmp     RES 1 ;temporary storage

time    RES 1 ;time counter
time1   RES 1 ;time counter
time2   RES 1 ;time counter
pos     RES 1 ;lut position
sncount RES 1 ;serial number counter
byte    RES 1 ;bytecounter
sn      RES 7 ;iButton serial number (no CRC)






;***** MEMORY START *****

  ORG 0x00    ;processor reset vector
  goto  MAIN  ;jump to main program


;***** INTERRUPT SERVICE ROUTINE *****

  ORG 0x04    ;interrupt vector location
  retfie


;************** MAIN **************


MAIN
;
; Initialise the ports
;

  bcf     STATUS, RP0   ;bank 0
  clrf    PORTA         ;clear output latches
  clrf    PORTB         ;clear output latches

  bsf     STATUS, RP0   ;bank 1
  movlw   128
  movwf   OPTION_REG    ;pull-ups off, interrupt on falling edge

  clrf    TRISA         ;all PORTA pins as outputs
  bsf     TRISA, IB     ;now iButton pin set to input
  clrf    TRISB         ;all PORTB pins as outputs
  bcf     STATUS, RP0   ;bank 0


;
; check for iButton presence
;
PRESENCE

  call    but_presence  ;check for iButton
  btfss   bdata, 0
  goto    PRESENCE      ;no iButton

  movlw   YELLOW        ;turn on red and green to indicate ID reading
  movwf   PORTB         ;do it

;
; Read iButton ID
;
READID

  clrf    crc       ;clear the CRC
  movlw   RROM
  call    but_wr    ;send Read ROM command

; read 7 byte serial number
  movlw   sn
  movwf   FSR       ;initialise pointer
  movlw   7         ;7 bytes
  movwf   byte      ;store

read
  call    but_rd    ;read ID byte
  movf    bdata, W  ;get ID byte
  call    calc_crc  ;update CRC
  movf    bdata, W  ;get ID byte
  movwf   INDF      ;write to indirect reg
  incf    FSR, F    ;increment pointer
  decfsz  byte, F   ;decrement byte, skip if zero
  goto    read      ;next byte

; read CRC
  call    but_rd    ;read CRC Checksum
  movf    crc, W    ;get the calculated CRC
  subwf   bdata, W  ;W=bdata-crc
  btfsc   STATUS, Z ;bit test, skip if clear
  goto    LOOKUP    ;receive ok


BADREAD
  movlw   3             ;3 loops
  movwf   byte          ;store

loopbr
  bcf     PORTB, GREEN  ;turn off green (red only)
  movlw   3
  call    delayx00ms    ; wait 3/10 s
  bsf     PORTB, GREEN  ;turn on green (yellow)
  movlw   3
  call    delayx00ms    ; wait 3/10 s
  decfsz  byte, F       ;decrement byte, skip if zero
  goto    loopbr        ;next loop

  movlw   BLACK         ;turn off red and green (and implicitly lock relay)
  movwf   PORTB         ;do it
  movlw   20
  call    delayx00ms    ;wait 2s
  goto    PRESENCE      ;try again


;
; Check against registered ID codes
;

LOOKUP
  movlw   BLACK         ;turn off red and green (and implicitly lock relay)
  movwf   PORTB         ;do it

;
;  lookup iButton's serial number
;
  clrf    pos           ;init lut pointer
  movlw   SID           ;get number of stored ID codes
  movwf   sncount       ;store
nextsn
  clrf    tmp
  movlw   sn            ;init ram pointer to read ID code
  movwf   FSR
  movlw   7             ;7 bytes to check
  movwf   byte          ;store
nxt
  movlw   TAB_PG        ;set page of
  movwf   PCLATH        ;table begin
  movf    pos, W
  call    LUT
  subwf   INDF, W       ;W=INDF-W
  btfss   STATUS, Z
  incf    tmp, F        ;if (w>0) inc(tmp)
  decfsz  byte, F
  goto    nextb         ;next byte
  movf    tmp, F
  btfsc   STATUS, Z
  goto    MATCH
  decf    sncount, F
  btfsc   STATUS, Z
  goto    NOMATCH
  incf    pos, F        ;next sn
  goto    nextsn
nextb
  incf    FSR, F
  incf    pos, F
  goto    nxt


;
; ID code did not match
;
NOMATCH
  bsf     PORTB, RED    ;turn on red
  movlw   200
  call    delayx00ms    ;wait 20s
  bcf     PORTB, RED    ;turn off red

  goto    PRESENCE      ;try again


;
; ID code is a match
;
MATCH
  bsf     PORTB, GREEN  ;turn on green
  bsf     PORTB, LOCK   ;activate lock relay
  movlw   20
  call    delayx00ms    ;wait 2s
  movlw   BLACK         ;turn off red, green and deactivate lock relay
  movwf   PORTB         ;do it

  goto    PRESENCE      ;try again


;
;  ibutton.inc - Dallas iButton read/write routines
;
; 09.11.98 - last changed: 09.11.98
; (c) M.Prinke

;
;  check if any iButton is present
;

but_presence

; reset pulse
  clrf    bdata
  bcf     PORTA, IB     ; make output latch low
  bsf     STATUS, RP0   ; Bank 1
  bcf     TRISA, IB     ; make pin output
  movlw   50
  call    delay10us     ; wait 500us
  bsf     TRISA, IB     ; make pin input
  bcf     STATUS, RP0   ; Bank 0
; check presence pulse
  movlw   9
  call    delay10us     ; wait 90us
  btfss   PORTA, IB     ; high->no button, bdata<0>=0
  bsf     bdata, 0      ; low->button present,
  movlw   50
  call    delay10us     ; wait 500us
  return                ; bdata<0>=1
;
;  write zero-bit to iButton
;
but_w0
  bcf     PORTA, IB     ; make output latch low
  bsf     STATUS, RP0   ; Bank 1
  bcf     TRISA, IB     ; make output pin (low)
  movlw   9
  call    delay10us     ; wait 90us
  bsf     TRISA, IB     ; make input pin
  bcf     STATUS, RP0   ; Bank 0
  movlw   3
  call    delay10us     ; wait 30us
  return
;
;  write one-bit to iButton
;
but_w1
  bcf     PORTA, IB     ; make output latch low
  bsf     STATUS, RP0   ; Bank 1
  bcf     TRISA, IB     ; make output pin (low)
  movlw   1
  call    delay10us     ; wait 10us
  bsf     TRISA, IB     ; make input pin
  bcf     STATUS, RP0   ; Bank 0
  movlw   11
  call    delay10us     ; wait 110us
  return
;
;  write byte to iButton - LSB first
;
but_wr
  movwf   bdata
  movlw   8
  movwf   bcount
sh
  rrf     bdata, F      ; shift byte
  btfss   STATUS, C     ;  to carry
  goto    w0
  call    but_w1        ; write one
  goto    nb
w0
  call    but_w0        ; write zero
nb
  decfsz  bcount, F     ; next bit?
  goto    sh
  return
;
;  read bit from iButton
;
but_r
  clrf    tmp
  bcf     PORTA, IB
  bsf     STATUS, RP0   ; Bank 1
  bcf     TRISA, IB     ; make output pin (low)
  movlw   1
  call    delay10us     ; wait 10us
  bsf     TRISA, IB     ; make input pin
  bcf     STATUS, RP0   ; Bank 0
  movlw   1
  call    delay10us     ; wait 10us
  btfsc   PORTA, IB     ; sample iButton
  bsf     tmp, 0
  rrf     tmp, W        ; shift bit into carry
  movlw   8
  call    delay10us     ; wait 80us
  return


;
;  read byte from iButton - LSB first
;
but_rd
  movlw   8
  movwf   bcount
rdb
  call    but_r
  rrf     bdata, F      ; shift carry to bdata
  decfsz  bcount, F     ; next bit?
  goto    rdb
  return                ; byte in bdata

;
;  calculate CRC, polynomial is x^8+x^5+x^4+1
;
calc_crc
  movwf   crcb          ; byte to update CRC with
  movlw   8
  movwf   bcount
  movf    crcb, W
crcl
  xorwf   crc, W        ; loop over 8 bits
  movwf   tmp
  rrf     tmp, F
  movf    crc, W
  btfsc   STATUS, C
  xorlw   0x18
crcz
  movwf   crc
  rrf     crc, F
  rrf     crcb, F
  bcf     crcb, 7
  btfsc   STATUS, C
  bsf     crcb, 7
  movf    crcb, W
  decfsz  bcount, F
  goto    crcl
  return

;
;  wait for <workingregister> times 10 microseconds
;
delay10us
  movwf   time
dly10us
  goto    $+1
  goto    $+1
  goto    $+1
  nop
  decfsz  time, F
  goto    dly10us
  return
;
;  wait for approx. 100ms
;
delay100ms
  movlw   100
  movwf   time1
dly100ms
  movlw   100
  call    delay10us
  decfsz  time1, F
  goto  dly100ms
  return
;
;  wait for <workingregister> times 100 ms
;
delayx00ms
  movwf   time2
dlyx00ms
  call    delay100ms
  decfsz  time2, F
  goto    dlyx00ms
  return

;
;  iButton serial number lookup table, LSB first
;
  ORG 0x0100

LUT
  addwf PCL, F
  retlw 0x01    ; SN button1
  retlw 0xBE
  retlw 0xC8
  retlw 0xE6
  retlw 0x06
  retlw 0x00
  retlw 0x00
  retlw 0x11    ; SN button2
  retlw 0x22
  retlw 0x33
  retlw 0x00
  retlw 0x00
  retlw 0x00
  retlw 0x00
  retlw 0x01    ; SN button3
  retlw 0xFC
  retlw 0xA9
  retlw 0xE6
  retlw 0x06
  retlw 0x00
  retlw 0x00
  end
